@@ -10,6 +10,7 @@
 #include <syslog.h>
 #include <signal.h>
 #include <wait.h>
+#include <sysfs/libsysfs.h>
 
 #include "devinfo.h"
 #include "checkers.h"
@@ -30,11 +31,13 @@
 #endif
 #define LOG(x, y, z...) if (DEBUG >= x) syslog(x, y, ##z)
 
+/* global */
+int from_sighup;
+
 struct path
 {
 	int major;
 	int minor;
-	char mapname[MAPNAMESIZE];
 	int (*checkfn) (char *);
 };
 
@@ -67,12 +70,38 @@ int makenode (char *devnode, int major, int minor)
 	dev = makedev (major, minor);
         unlink (devnode);
 	
-	return mknod(devnode, S_IFBLK | S_IRUSR | S_IWUSR, dev);
+	return mknod(devnode, S_IFCHR | S_IRUSR | S_IWUSR, dev);
 }
 
-int select_checkfn(struct path *path_p)
+static int
+blacklist (char * dev) {
+	int i;
+	static struct {
+		char * headstr;
+		int lengh;
+	} blist[] = {
+		{"cciss", 5},
+		{"fd", 2},
+		{"hd", 2},
+		{"md", 2},
+		{"dm", 2},
+		{"sr", 2},
+		{"scd", 3},
+		{"ram", 3},
+		{"raw", 3},
+		{"loop", 4},
+		{NULL, 0},
+	};
+
+	for (i = 0; blist[i].lengh; i++) {
+		if (strncmp (dev, blist[i].headstr, blist[i].lengh) == 0)
+                        return 1;
+	}
+	return 0;
+}
+
+int select_checkfn(struct path *path_p, char *devname)
 {
-	char devnode[FILENAMESIZE];
 	char vendor[8];
 	char product[16];
 	char rev[4];
@@ -81,35 +110,19 @@ int select_checkfn(struct path *path_p)
 	/* default checkfn */
 	path_p->checkfn = &readsector0;
 	
-	sprintf (devnode, "/tmp/.select.%i.%i", path_p->major, path_p->minor);
-
-	r = makenode (devnode, path_p->major, path_p->minor);
-	
-	if (r < 0) {
-		LOG(2, "[select_checkfn] can not make node %s", devnode);
-		return r;
-	}
-
-	r = get_lun_strings(vendor, product, rev, devnode);
+	r = get_lun_strings(vendor, product, rev, devname);
 
 	if (r) {
 		LOG(2, "[select_checkfn] can not get strings");
 		return r;
 	}
 
-	r = unlink (devnode);
-
-	if (r < 0) {
-		LOG(2, "[select_checkfn] can not unlink %s", devnode);
-		return r;
-	}
-
 	static struct {
 		char * vendor;
 		char * product;
 		int (*checkfn) (char *);
 	} wlist[] = {
-		{"COMPAQ  ", "HSV110 (C)COMPAQ", &tur},
+		{"COMPAQ  ", "HSV110 (C)COMPAQ", &readsector0},
 		{"COMPAQ  ", "MSA1000         ", &tur},
 		{"COMPAQ  ", "MSA1000 VOLUME  ", &tur},
 		{"DEC     ", "HSG80           ", &tur},
@@ -121,6 +134,8 @@ int select_checkfn(struct path *path_p)
 		if (strncmp(vendor, wlist[i].vendor, 8) == 0 &&
 		    strncmp(product, wlist[i].product, 16) == 0) {
 			path_p->checkfn = wlist[i].checkfn;
+			LOG (2, "[select_checkfn] set checkfn for %s",
+			     devname);
 		}
 	}
 
@@ -229,15 +244,8 @@ int checkpath (struct path *path_p)
 	char devnode[FILENAMESIZE];
 	int r;
 	
-	LOG (2, "[checkpath] checking path %i:%i", path_p->major, path_p->minor);
 	sprintf (devnode, "/tmp/.checker.%i.%i", path_p->major, path_p->minor);
 	
-	if (path_p->checkfn == NULL) {
-		LOG (1, "[checkpath] test function not set for path %i:%i",
-		     path_p->major, path_p->minor);
-		return 1;
-	}
-
 	r = makenode (devnode, path_p->major, path_p->minor);
 
 	if (r < 0) {
@@ -248,118 +256,88 @@ int checkpath (struct path *path_p)
 	r = path_p->checkfn(devnode);
 	unlink (devnode);
 				
+	LOG (2, "[checkpath] checked path %i:%i => %i",
+	     path_p->major, path_p->minor, r);
+
 	return r;
 }
-		
+
 int updatepaths (struct devmap *devmaps, struct paths *failedpaths)
 {
 	struct path *path_p;
-	struct devmap *devmaps_p;
-	void *next;
-	struct dm_task *dmt;
-	uint64_t start, length;
-	char *target_type = NULL;
-	char *params, *p1, *p2;
-	char word[6];
-	int i;
+	struct sysfs_directory * sdir;
+	struct sysfs_directory * devp;
+	char sysfs_path[FILENAMESIZE];
+	char path[FILENAMESIZE];
+	char attr_path[FILENAMESIZE];
+	char attr_buff[17];
+	char *p1, *p2;
+	char word[5];
 	
-	path_p = failedpaths->paths_h;
-	
-	pthread_mutex_lock (failedpaths->lock);
-	memset (failedpaths->paths_h, 0, MAXPATHS * sizeof (struct path));
-
-	/* first pass */
-	/* ask DM the failed path list */
-
-	devmaps_p = devmaps;
-
-	while (*devmaps_p->mapname != 0x0) {
-		next = NULL;
-		
-		if (!(dmt = dm_task_create(DM_DEVICE_STATUS)))
-			break;
-		
-		if (!dm_task_set_name(dmt, devmaps_p->mapname))
-			goto out;
-		
-		if (!dm_task_run(dmt))
-			goto out;
-
-		do {
-			next = dm_get_next_target(dmt, next, &start, &length,
-						   &target_type, &params);
-
-			/* begin ugly parser */
-			p1 = params;
-			p2 = params;
-			while (*p1) {
-				/* p2 lags at the begining of the word p1 parses */
-				while (*p1 != ' ') {
-					/* if the current word is a path */
-					if (*p1 == ':') {
-						/* p1 jumps to path state */
-						while (*p1 != 'A' && *p1 != 'F')
-							p1++; 
-						
-						/* store path info */
-
-						path_p->checkfn = NULL;
-
-						i = 0;
-						memset (&word, 'O', 6 * sizeof (char));
-						while (*p2 != ':') {
-							word[i++] = *p2;
-							p2++;
-						}
-						path_p->major = atoi (word);
-						
-						p2++;
-						i = 0;
-						memset (&word, 'O', 6 * sizeof (char));
-						while (*p2 != ' ') {
-							word[i++] = *p2;
-							p2++;
-						}
-						path_p->minor = atoi (word);
-
-						strcpy (path_p->mapname, devmaps_p->mapname);
-
-						/* 
-						 * discard active paths
-						 * don't trust the A status flag : double check
-						 */
-						if (*p1 == 'A' &&
-						    !select_checkfn (path_p) &&
-						    checkpath (path_p)) {
-							LOG(2, "[updatepaths] discard %i:%i as valid path",
-							    path_p->major, path_p->minor);
-							p1++;
-							memset (path_p, 0, sizeof(struct path));
-							continue;
-						}
-						
-						path_p++;
-
-						/* test vector overflow */
-						if (path_p - failedpaths->paths_h >= MAXPATHS * sizeof (struct path)) {
-							LOG (1, "[updatepaths] path_h overflow");
-							pthread_mutex_unlock (failedpaths->lock);
-							return 1;
-						}
-					}
-					p1++;
-				}
-				p2 = p1;
-				p1++;
-			}
-		} while (next);
-			
-out:
-		dm_task_destroy(dmt);
-		devmaps_p++;
-		
+	if (sysfs_get_mnt_path (sysfs_path, FILENAMESIZE)) {
+		LOG (2, "[updatepaths] can not find sysfs mount point");
+		return 1;
 	}
 
+	sprintf (path, "%s/block", sysfs_path);
+	sdir = sysfs_open_directory (path);
+	sysfs_read_directory (sdir);
+
+	pthread_mutex_lock (failedpaths->lock);
+	memset (failedpaths->paths_h, 0, MAXPATHS * sizeof (struct path));
+	path_p = failedpaths->paths_h;
+
+	dlist_for_each_data (sdir->subdirs, devp, struct sysfs_directory) {
+		if (blacklist (devp->name)) {
+			LOG (2, "[updatepaths] %s blacklisted", devp->name);
+			continue;
+		}
+
+		sysfs_read_directory (devp);
+
+		if (devp->links == NULL)
+			continue;
+
+		sprintf(attr_path, "%s/block/%s/device/generic/dev",
+			sysfs_path, devp->name);
+
+		if (0 > sysfs_read_attribute_value(attr_path, attr_buff, 11))
+			return 1;
+
+		p1 = &word[0];
+		p2 = &attr_buff[0];
+		memset (&word, 0, 5 * sizeof (char));
+		
+		while (*p2 != ':') {
+			*p1 = *p2;
+			p1++;
+			p2++;
+		}
+		path_p->major = atoi (word);
+
+		p1 = &word[0];
+		p2++;
+		memset (&word, 0, 5 * sizeof (char));
+
+		while (*p2 != ' ') {
+			*p1 = *p2;
+			p1++;
+			p2++;
+		}
+		path_p->minor = atoi (word);
+
+		if (!select_checkfn (path_p, devp->name) &&
+		    checkpath (path_p)) {
+			LOG(2, "[updatepaths] discard %i:%i as valid path",
+			    path_p->major, path_p->minor);
+			memset (path_p, 0, sizeof(struct path));
+			continue;
+		}
+
+		LOG (2, "[updatepaths] %i:%i added to failedpaths",
+		     path_p->major, path_p->minor);
+		path_p++;
+	}
 	pthread_mutex_unlock (failedpaths->lock);
 	return 0;
 }
@@ -437,6 +415,8 @@ void *waiterloop (void *ap)
 	struct event_thread *waiters, *waiters_p;
 	pthread_attr_t attr;
 	int r;
+	char *cmdargs[4] = {MULTIPATH, "-q", "-S"};
+	int status;
 
 	/* inits */
 	failedpaths = (struct paths *)ap;
@@ -447,17 +427,28 @@ void *waiterloop (void *ap)
 	pthread_attr_setstacksize (&attr, 32 * 1024);
 
 	while (1) {
+		/* upon event and initial startup, do a preliminary
+		   multipath exec, no signal to avoid recursion.
+		   don't run multipath if we are waked from SIGHUP
+		   because it already ran */
+		if (!from_sighup) {
+			LOG (1, "[waiterloop] exec multipath");
+			if (fork () == 0)
+				execve (cmdargs[0], cmdargs, NULL);
+			wait (&status);
+		} else
+			from_sighup = 0;
 		
 		/* update devmap list */
-		LOG (1, "[event thread] refresh devmaps list");
+		LOG (1, "[waiterloop] refresh devmaps list");
 		get_devmaps (devmaps);
 
 		/* update failed paths list */
-		LOG (1, "[event thread] refresh failpaths list");
+		LOG (1, "[waiterloop] refresh failpaths list");
 		updatepaths (devmaps, failedpaths);
 		
 		/* start waiters on all devmaps */
-		LOG (1, "[event thread] start up event loops");
+		LOG (1, "[waiterloop] start up event loops");
 		waiters_p = waiters;
 		devmaps_p = devmaps;
 
@@ -489,7 +480,7 @@ void *waiterloop (void *ap)
 				pthread_mutex_unlock (waiters_p->waiter_lock);
 			}
 			
-			LOG (1, "[event thread] create event thread for %s", waiters_p->mapname);
+			LOG (1, "[waiterloop] create event thread for %s", waiters_p->mapname);
 			pthread_create (waiters_p->thread, &attr, waitevent, waiters_p);
 			pthread_detach (*waiters_p->thread);
 out:
@@ -502,7 +493,7 @@ out:
 		pthread_cond_wait(event, event_lock);
 		pthread_mutex_unlock (event_lock);
 
-		LOG (1, "[event thread] event caught");
+		LOG (1, "[waiterloop] event caught");
 	}
 
 	return (NULL);
@@ -512,7 +503,7 @@ void *checkerloop (void *ap)
 {
 	struct paths *failedpaths;
 	struct path *path_p;
-	char *cmdargs[4] = {MULTIPATH, "-D", NULL, NULL};
+	char *cmdargs[5] = {MULTIPATH, "-D", NULL, NULL, "-q"};
 	char major[5];
 	char minor[5];
 	int status;
@@ -528,7 +519,8 @@ void *checkerloop (void *ap)
 		while (path_p->major != 0) {
 			
 			if (checkpath (path_p)) {
-				LOG (1, "[checker thread] reconfigure %s\n", path_p->mapname);
+				LOG (1, "[checker thread] exec multipath for device %i:%i\n",
+				     path_p->major, path_p->minor);
 				snprintf (major, 5, "%i", path_p->major);
 				snprintf (minor, 5, "%i", path_p->minor);
 				cmdargs[2] = major;
@@ -537,7 +529,7 @@ void *checkerloop (void *ap)
 					execve (cmdargs[0], cmdargs, NULL);
 
 				wait (&status);
-				/* MULTIPATH will ask for failedpaths refresh (SIGHUP) */
+				/* MULTIPATH will send back a SIGHUP */
 			}
 			
 			path_p++;
@@ -550,7 +542,7 @@ void *checkerloop (void *ap)
 			}
 		}
 		pthread_mutex_unlock (failedpaths->lock);
-		sleep(CHECKINT);
+		sleep (CHECKINT);
 	}
 
 	return (NULL);
@@ -622,6 +614,9 @@ void sighup (int sig)
 {
 	LOG (1, "[master thread] SIGHUP caught : refresh devmap list");
 
+	/* signal updatepaths() that we come from SIGHUP */
+	from_sighup = 1;
+
 	/* ask for failedpaths refresh */
 	pthread_mutex_lock (event_lock);
 	pthread_cond_signal(event);
